/*
 * LMDB Abstraction Layer.
 * Copyright (C) 2023  Yury Gubich <blue@macaw.me>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "storagecommon.h"

#include "cursorcommon.h"

#define UNUSED(x) (void)(x)

/**
 * \class LMDBAL::iStorage
 *
 * \brief Storage interface
 *
 * This is a interface-like class, it's designed to be an inner database interface to
 * be used as a polymorphic entity, and provide protected interaction with the database
 * from the heirs code
 */

/**
 * \brief Constructs a storage interface
 *
 * \param[in]   parent      - LMDBAL::Base pointer for the owning database (borrowed)
 * \param[in]   name        - the name of the storage
 * \param[in]   duplicates  - true if key duplicates are allowed (false by default)
 */
LMDBAL::StorageCommon::StorageCommon(Base* parent, const std::string& name, bool duplicates):
    dbi(),
    db(parent),
    name(name),
    duplicates(duplicates)
{}

/**
 * \brief Destroys a storage interface
 */
LMDBAL::StorageCommon::~StorageCommon () {}

/**
 * \brief A private virtual function to close each storage in the database
 */
void LMDBAL::StorageCommon::close() {
    mdb_dbi_close(db->environment, dbi);
}

/**
 * \brief Checks if the transaction is still active, returns inner TransactionID
 *
 * This method is for internal usage only
 *
 * \param[in] txn       - a transaction, you want to extract ID from
 * \param[in] action    - a description of what you're going to do, in case of exception the error description will be verbose
 * \returns             - Transaction inner TransactionID
 *
 * \exception LMDBAL::TransactionTerminated     thrown if the passed transaction not active, any action with it's inner ID is an error
 */
LMDBAL::TransactionID LMDBAL::StorageCommon::extractTransactionId(const Transaction& txn, const std::string& action) const {
    if (!txn.isActive())
        throw TransactionTerminated(db->name, name, action);

    return txn.txn;
}

/**
 * \brief Drops content of a storage interface
 *
 * Designed to drop storage content
 *
 * \exception LMDBAL::Closed    thrown if the database was closed
 * \exception LMDBAL::Unknown   thrown if something unexpected happened
 */
void LMDBAL::StorageCommon::drop() {
    ensureOpened(dropMethodName);

    TransactionID txn = beginTransaction();
    int rc = StorageCommon::drop(txn);
    if (rc != MDB_SUCCESS) {
        abortTransaction(txn);
        throw Unknown(db->name, mdb_strerror(rc), name);
    }

    db->commitTransaction(txn);
    handleDrop();
}

/**
 * \brief Drops content of a storage interface (transaction variant)
 *
 * Just performs content drop
 *
 * \param[in] transaction   - transaction ID, must be writable transaction!
 * \returns MDB_SUCCESS if everything went fine, MDB_<error> code otherwise
 */
int LMDBAL::StorageCommon::drop(TransactionID transaction) {
    return mdb_drop(transaction, dbi, 0);
}

/**
 * \brief Drops content of a storage interface (public transaction variant)
 *
 * Just performs content drop
 *
 * \param[in] txn   - transaction ID, must be writable transaction!
 * \returns MDB_SUCCESS if everything went fine, MDB_<error> code otherwise
 *
 * \exception LMDBAL::TransactionTerminated   thrown if the transaction was not active
 */
int LMDBAL::StorageCommon::drop(const WriteTransaction& txn) {
    ensureOpened(dropMethodName);
    return drop(extractTransactionId(txn, dropMethodName));
}

/**
 * \brief Helper function, thows exception if the database is not opened
 *
 * \param[in] methodName    - name of the method this function is called from, just for display in std::exception::what() message
 *
 * \exception LMDBAL::Closed    thrown if the database was closed
 */
void LMDBAL::StorageCommon::ensureOpened(const std::string& methodName) const {
    if (!isDBOpened())
        throw Closed(methodName, db->name, name);
}

/**
 * \brief Storage size
 *
 * \returns amount of records in the storage
 *
 * \exception LMDBAL::Closed    thrown if the database was closed
 * \exception LMDBAL::Unknown   thrown if something unexpected happened
 */
LMDBAL::SizeType LMDBAL::StorageCommon::count() const {
    ensureOpened(countMethodName);

    TransactionID txn = beginReadOnlyTransaction();
    SizeType amount;
    try {
        amount = count(txn);
    } catch (...) {
        abortTransaction(txn);
        throw;
    }
    abortTransaction(txn);

    return amount;
}

/**
 * \brief Links cursor to the transaction it has been opened with
 *
 * This a service function is designed to be called by from LMDBAL::Cursor
 * Cursor must be opened by a public transaction
 *
 * \exception std::out_of_range     thrown if LMDBAL::Transaction pointer wasn't found in the LMDBAL::Base
 */

void LMDBAL::StorageCommon::attachCursorToTransaction (uint32_t cursorId, MDB_cursor* cursorHandle, CursorCommon* pointer) const {
    TransactionID txnID = _mdbCursorTxn(cursorHandle);
    Transaction* txn = db->transactions.at(txnID);
    txn->cursors[cursorId] = pointer;
}


/**
 * \brief Opens a transaction that is ment to be private to LMDBAL::Cursor, but public to the LMDBAL::Storage
 *
 * This method is ment to be called from LMDBAL::Cursor
 *
 * \param[out] cursor   - cursor handle
 * \param[in]  renew    - true if instead of opening cursor should be renewed
 *
 * \exception LMDBAL::Closed        thrown if you try to open the cursor on a closed database
 * \exception LMDBAL::Unknown       thrown if there was a problem opening the cursor by the lmdb, or to begin a transaction
 */
void LMDBAL::StorageCommon::openCursorTransaction (MDB_cursor** cursor, bool renew) const {
    ensureOpened(openCursorTransactionMethodName);

    TransactionID txn = beginReadOnlyTransaction();
    int result;
    if (renew)
        result = _mdbCursorRenew(txn, *cursor);
    else
        result = _mdbCursorOpen(txn, cursor);

    if (result != MDB_SUCCESS)
        throwUnknown(result, txn);

    transactionStarted(txn, true);
}

/**
 * \brief Closes transaction that is private to LMDBAL::Cursor, but public to the LMDBAL::Storage
 *
 * This method is ment to be called from LMDBAL::Cursor
 *
 * \param[in] cursor        - cursor handle
 * \param[in] closeCursor   - true if the cursor should also get closed, false if you wish to leave it open
 */
void LMDBAL::StorageCommon::closeCursorTransaction (MDB_cursor* cursor, bool closeCursor) const {
    TransactionID txn = _mdbCursorTxn(cursor);
    if (closeCursor)
        _mdbCursorClose(cursor);

    abortTransaction(txn);
    transactionAborted(txn);
}

/**
 * \brief Disconnects cursor from the transaction it has been opened with
 *
 * This a service function is designed to be called by from LMDBAL::Cursor
 * Cursor must be still opened by a public transaction
 *
 * \exception std::out_of_range     thrown if LMDBAL::Transaction pointer wasn't found in the LMDBAL::Base
 */

void LMDBAL::StorageCommon::disconnectCursorFromTransaction (uint32_t cursorId, MDB_cursor* cursorHandle) const {
    TransactionID txnID = _mdbCursorTxn(cursorHandle);
    Transaction* txn = db->transactions.at(txnID);
    txn->cursors.erase(cursorId);
}


/**
 * \brief Storage size (private transaction variant)
 *
 * \param[in] txn   - transaction ID, can be read-only transaction
 * \returns amount of records in the storage
 *
 * \exception LMDBAL::Unknown   thrown if something unexpected happened
 */
LMDBAL::SizeType LMDBAL::StorageCommon::count(TransactionID txn) const {
    MDB_stat stat;
    int rc = mdb_stat(txn, dbi, &stat);
    if (rc != MDB_SUCCESS)
        throw Unknown(db->name, mdb_strerror(rc), name);

    return stat.ms_entries;
}

/**
 * \brief Storage size (public transaction variant)
 *
 * \param[in] txn   - transaction, can be read-only transaction
 * \returns amount of records in the storage
 *
 * \exception LMDBAL::Closed                    thrown if the database was closed
 * \exception LMDBAL::Unknown                   thrown if something unexpected happened
 * \exception LMDBAL::TransactionTerminated     thrown if the passed transaction not active, any action with it's inner ID is an error
 */
LMDBAL::SizeType LMDBAL::StorageCommon::count(const Transaction& txn) const {
    ensureOpened(countMethodName);
    return count(extractTransactionId(txn, countMethodName));
}

/**
 * \brief Throws LMDBAL::Exist or LMDBAL::Unknown (transaction vairiant)
 *
 * Helper function ment to be used in heirs and reduce the code a bit
 *
 * \param[in] rc    - result of lmdb low level operation
 * \param[in] txn   - transaction ID to be aborted, any transaction
 * \param[in] key   - requested key string representation, just to show in std::exception::what() message
 *
 * \exception LMDBAL::Exist     thrown if rc == MDB_KEYEXIST
 * \exception LMDBAL::Unknown   thrown if rc != MDB_KEYEXIST
 */
void LMDBAL::StorageCommon::throwDuplicateOrUnknown(int rc, TransactionID txn, const std::string& key) const {
    abortTransaction(txn);
    throwDuplicateOrUnknown(rc, key);
}

/**
 * \brief Throws LMDBAL::NotFound or LMDBAL::Unknown (transaction vairiant)
 *
 * Helper function ment to be used in heirs and reduce the code a bit
 *
 * \param[in] rc    - result of lmdb low level operation
 * \param[in] txn   - transaction ID to be aborted, any transaction
 * \param[in] key   - requested key string representation, just to show in std::exception::what() message
 *
 * \exception LMDBAL::NotFound  thrown if rc == MDB_NOTFOUND
 * \exception LMDBAL::Unknown   thrown if rc != MDB_NOTFOUND
 */
void LMDBAL::StorageCommon::throwNotFoundOrUnknown(int rc, LMDBAL::TransactionID txn, const std::string& key) const {
    abortTransaction(txn);
    throwNotFoundOrUnknown(rc, key);
}

/**
 * \brief Throws LMDBAL::Exist or LMDBAL::Unknown
 *
 * Helper function ment to be used in heirs and reduce the code a bit
 *
 * \param[in] rc    - result of lmdb low level operation
 * \param[in] key   - requested key string representation, just to show in std::exception::what() message
 *
 * \exception LMDBAL::Exist     thrown if rc == MDB_KEYEXIST
 * \exception LMDBAL::Unknown   thrown if rc != MDB_KEYEXIST
 */
void LMDBAL::StorageCommon::throwDuplicateOrUnknown(int rc, const std::string& key) const {
    if (rc == MDB_KEYEXIST)
        throwDuplicate(key);
    else
        throwUnknown(rc);
}

/**
 * \brief Throws LMDBAL::NotFound or LMDBAL::Unknown (transaction variant)
 *
 * Helper function ment to be used in heirs and reduce the code a bit
 *
 * \param[in] rc    - result of lmdb low level operation
 * \param[in] key   - requested key string representation, just to show in std::exception::what() message
 *
 * \exception LMDBAL::NotFound  thrown if rc == MDB_NOTFOUND
 * \exception LMDBAL::Unknown   thrown if rc != MDB_NOTFOUND
 */
void LMDBAL::StorageCommon::throwNotFoundOrUnknown(int rc, const std::string& key) const {
    if (rc == MDB_NOTFOUND)
        throwNotFound(key);
    else
        throwUnknown(rc);
}

/**
 * \brief Throws LMDBAL::Unknown (transaction vairiant)
 *
 * Helper function ment to be used in heirs and reduce the code a bit
 *
 * \param[in] rc    - result of lmdb low level operation
 * \param[in] txn   - transaction ID to be aborted, any transaction
 *
 * \exception LMDBAL::Unknown   thrown everytime
 */
void LMDBAL::StorageCommon::throwUnknown(int rc, LMDBAL::TransactionID txn) const {
    abortTransaction(txn);
    throwUnknown(rc);
}

/**
 * \brief Database name
 *
 * Ment to be used in heirs, to provide some sort of interface to acces to some of the database information
 *
 * \returns database name
 */
const std::string & LMDBAL::StorageCommon::dbName() const {
    return db->name;}

/**
 * \brief Is database opened
 *
 * Ment to be used in heirs, to provide some sort of interface to acces to some of the database information
 *
 * \returns true if database is ipened, false otherwise
 */
bool LMDBAL::StorageCommon::isDBOpened() const {
    return db->opened;}

/**
 * \brief Throws LMDBAL::Unknown
 *
 * Helper function ment to be used in heirs and reduce the code a bit
 *
 * \param[in] rc    - result of lmdb low level operation
 *
 * \exception LMDBAL::Unknown   thrown everytime
 */
void LMDBAL::StorageCommon::throwUnknown(int rc) const {
    throw Unknown(db->name, mdb_strerror(rc), name);}

/**
 * \brief Throws LMDBAL::Unknown
 *
 * Helper function ment to be used in heirs and reduce the code a bit
 *
 * \param[in] message    - a message you wish to appear in the exception reason
 *
 * \exception LMDBAL::Unknown   thrown everytime
 */
void LMDBAL::StorageCommon::throwUnknown(const std::string& message) const {
    throw Unknown(db->name, message, name);}

/**
 * \brief Throws LMDBAL::Exist
 *
 * Helper function ment to be used in heirs and reduce the code a bit
 *
 * \param[in] key   - requested key string representation, just to show in std::exception::what() message
 *
 * \exception LMDBAL::Exist   thrown everytime
 */
void LMDBAL::StorageCommon::throwDuplicate(const std::string& key) const {
    throw Exist(key, db->name, name);}

/**
 * \brief Throws LMDBAL::NotFound
 *
 * Helper function ment to be used in heirs and reduce the code a bit
 *
 * \param[in] key   - requested key string representation, just to show in std::exception::what() message
 *
 * \exception LMDBAL::NotFound   thrown everytime
 */
void LMDBAL::StorageCommon::throwNotFound(const std::string& key) const {
    throw NotFound(key, db->name, name);}

/**
 * \brief Throws LMDBAL::CursorNotReady
 *
 * Helper function ment to be used in heirs and reduce the code a bit
 *
 * \param[in] method   - called cursor method name, just to show in std::exception::what() message
 *
 * \exception LMDBAL::CursorNotReady   thrown everytime
 */
void LMDBAL::StorageCommon::throwCursorNotReady(const std::string& method) const {
    throw CursorNotReady(method, db->name, name);}

/**
 * \brief Begins read-only transaction
 *
 * Ment to be called from heirs, name is reported to the database just to be displayed in std::exception::what() message
 *
 * \returns read only transaction
 */
LMDBAL::TransactionID LMDBAL::StorageCommon::beginReadOnlyTransaction() const {
    return db->beginPrivateReadOnlyTransaction(name);}

/**
 * \brief Begins writable transaction
 *
 * Ment to be called from heirs, name is reported to the database just to be displayed in std::exception::what() message
 *
 * \returns read only transaction
 */
LMDBAL::TransactionID LMDBAL::StorageCommon::beginTransaction() const {
    return db->beginPrivateTransaction(name);}

/**
 * \brief Aborts transaction
 *
 * Ment to be called from heirs, name is reported to the database just to be displayed in std::exception::what() message
 */
void LMDBAL::StorageCommon::abortTransaction(LMDBAL::TransactionID id) const {
    db->abortPrivateTransaction(id, name);}

/**
 * \brief Commits transaction
 *
 * Ment to be called from heirs, name is reported to the database just to be displayed in std::exception::what() message
 *
 * \exception LMDBAL::Unknown   thrown if something unexpected happened
 */
void LMDBAL::StorageCommon::commitTransaction(LMDBAL::TransactionID id) {
    db->commitPrivateTransaction(id, name);}

/**
 * \brief called on beginning of public transaction
 *
 * This function is called on every storage of the database
 * when user calls LMDBAL::Base::beginTransaction() or LMDBAL::Base::beginReadOnlyTransaction()
 *
 * This function is met to be reimplemented in heirs
 * if the heir code requires some transaction custom handling
 *
 * \param[in] txn       - ID of started transaction
 * \param[in] readOnly  - true if transaction is read-only, false otherwise
 */
void LMDBAL::StorageCommon::transactionStarted(LMDBAL::TransactionID txn, bool readOnly) const {
    UNUSED(txn);
    UNUSED(readOnly);
}

/**
 * \brief called on commitment of public transaction
 *
 * This function is called on every storage of the database
 * when user calls LMDBAL::Base::commitTransaction(LMDBAL::TransactionID)
 *
 * This function is met to be reimplemented in heirs
 * if the heir code requires some transaction custom handling
 *
 * \param[in] txn       - ID of started transaction
 */
void LMDBAL::StorageCommon::transactionCommited(LMDBAL::TransactionID txn) {
    UNUSED(txn);}

/**
 * \brief called on abortion of public transaction
 *
 * This function is called on every storage of the database
 * when user calls LMDBAL::Base::abortTransaction(LMDBAL::TransactionID)
 *
 * This function is met to be reimplemented in heirs
 * if the heir code requires some transaction custom handling
 *
 * \param[in] txn       - ID of started transaction
 */
void LMDBAL::StorageCommon::transactionAborted(LMDBAL::TransactionID txn) const {
    UNUSED(txn);}

/**
 * \brief A method where database additionally handles drop
 *
 * It's a protected method that is called to optimise drop process
 * after the transaction is commited. Used just for optimisations.
 */
void LMDBAL::StorageCommon::handleDrop() {}

int LMDBAL::StorageCommon::_mdbOpen(MDB_txn *txn, unsigned int flags) {
    return mdb_dbi_open(txn, name.c_str(), flags, &dbi);
}

int LMDBAL::StorageCommon::_mdbPut(MDB_txn* txn, MDB_val& key, MDB_val& data, unsigned int flags) {
    return mdb_put(txn, dbi, &key, &data, flags);
}

int LMDBAL::StorageCommon::_mdbGet(MDB_txn* txn, MDB_val& key, MDB_val& data) const {
    return mdb_get(txn, dbi, &key, &data);
}

int LMDBAL::StorageCommon::_mdbDel(MDB_txn* txn, MDB_val& key) {
    return mdb_del(txn, dbi, &key, NULL);
}

int LMDBAL::StorageCommon::_mdbDel(MDB_txn* txn, MDB_val& key, MDB_val& data) {
    return mdb_del(txn, dbi, &key, &data);
}

int LMDBAL::StorageCommon::_mdbStat(MDB_txn* txn, MDB_stat& stat) const {
    return mdb_stat(txn, dbi, &stat);
}

int LMDBAL::StorageCommon::_mdbFlags(MDB_txn* txn, uint32_t& flags) const {
    return mdb_dbi_flags(txn, dbi, &flags);
}


int LMDBAL::StorageCommon::_mdbCursorOpen(MDB_txn* txn, MDB_cursor** cursor) const {
    return mdb_cursor_open(txn, dbi, cursor);
}

void LMDBAL::StorageCommon::_mdbCursorClose(MDB_cursor *cursor) const {
    mdb_cursor_close(cursor);
}

int LMDBAL::StorageCommon::_mdbCursorGet(MDB_cursor* cursor, MDB_val& key, MDB_val& data, MDB_cursor_op operation) const {
    return mdb_cursor_get(cursor, &key, &data, operation);
}

int LMDBAL::StorageCommon::_mdbCursorSet(MDB_cursor* cursor, MDB_val& key) const {
    return mdb_cursor_get(cursor, &key, NULL, MDB_SET);
}

int LMDBAL::StorageCommon::_mdbCursorDel(MDB_cursor* cursor, unsigned int flags) {
    return mdb_cursor_del(cursor, flags);
}

int LMDBAL::StorageCommon::_mdbCursorPut(MDB_cursor* cursor, MDB_val& key, MDB_val& data, unsigned int flags) {
    return mdb_cursor_put(cursor, &key, &data, flags);
}

int LMDBAL::StorageCommon::_mdbCursorRenew(MDB_txn* txn, MDB_cursor* cursor) const {
    return mdb_cursor_renew(txn, cursor);
}

MDB_txn* LMDBAL::StorageCommon::_mdbCursorTxn(MDB_cursor* cursor) const {
    return mdb_cursor_txn(cursor);
}
